use std::os::unix::io::AsRawFd;

const ETH_P_ALL: libc::c_short = 0x0003;
const SIOCGIFINDEX: libc::c_ulong = 0x8933;
const SIOCGIFMTU: libc::c_ulong = 0x8921;
const SIOCSIFMTU: libc::c_ulong = 0x8922;
const SIOCGIFHWADDR: libc::c_ulong = 0x8927;

#[repr(C)]
#[derive(Clone, Copy)]
#[allow(non_camel_case_types)]
union ifreq_data {
    ifr_mtu: libc::c_int,
    mac_addr: libc::sockaddr,
}

#[repr(C)]
#[allow(non_camel_case_types)]
struct ifreq {
    ifr_name: [libc::c_char; libc::IF_NAMESIZE],
    ifr_data: ifreq_data,
}

impl ifreq {
    #[cfg(target_os = "linux")]
    fn ifreq_for(interface: &str) -> ifreq {
        let mut ifreq = ifreq {
            ifr_name: [0; libc::IF_NAMESIZE],
            ifr_data: ifreq_data { ifr_mtu: 0 },
        };
        for (i, byte) in interface.as_bytes().iter().enumerate() {
            ifreq.ifr_name[i] = *byte as libc::c_char
        }
        ifreq
    }

    #[cfg(target_os = "linux")]
    fn ioctl(&mut self, lower: libc::c_int, cmd: libc::c_ulong) -> std::io::Result<ifreq_data> {
        unsafe {
            if libc::ioctl(lower, cmd as _, self as *mut Self) < 0 {
                return Err(std::io::Error::last_os_error());
            }
        }
        Ok(self.ifr_data)
    }
}

pub struct NetDevice {
    fd: libc::c_int,
    ifreq: ifreq,
    mac_addr: [u8; 6],
}

impl AsRawFd for NetDevice {
    fn as_raw_fd(&self) -> std::os::unix::io::RawFd {
        self.fd
    }
}

impl NetDevice {
    pub fn new(interface: &str) -> std::io::Result<Self> {
        #[cfg(target_os = "linux")]
        {
            let lower = unsafe {
                let lower = libc::socket(
                    libc::AF_PACKET,
                    libc::SOCK_RAW | libc::SOCK_NONBLOCK,
                    ETH_P_ALL.to_be() as i32,
                );
                if lower == -1 {
                    return Err(std::io::Error::last_os_error());
                }
                lower
            };

            let mut ifreq = ifreq::ifreq_for(interface);

            let ifreq_mac_addr = unsafe { ifreq.ioctl(lower, SIOCGIFHWADDR)?.mac_addr };
            let mut mac_addr = [0u8; 6];
            for i in 0..6 {
                mac_addr[i] = ifreq_mac_addr.sa_data[i] as u8;
            }

            println!(
                "Device MAC: {:02x}:{:02x}:{:02x}:{:02x}:{:02x}:{:02x}",
                mac_addr[0], mac_addr[1], mac_addr[2], mac_addr[3], mac_addr[4], mac_addr[5]
            );

            let mut dev = Self {
                fd: lower,
                ifreq,
                mac_addr,
            };

            dev.bind_interface()?;

            let mtu = dev.interface_mtu()?;
            println!("DEVICE MTU: {}", mtu);

            Ok(dev)
        }
        #[cfg(not(target_os = "linux"))]
        {
            Err(std::io::Error::new(
                std::io::ErrorKind::Other,
                "Not supported",
            ))
        }
    }

    pub fn mac_addr(&self) -> [u8; 6] {
        self.mac_addr
    }

    pub fn bind_interface(&mut self) -> std::io::Result<()> {
        #[cfg(target_os = "linux")]
        {
            let sockaddr = libc::sockaddr_ll {
                sll_family: libc::AF_PACKET as u16,
                sll_protocol: ETH_P_ALL.to_be() as u16,
                sll_ifindex: unsafe {
                    self.ifreq.ioctl(self.fd, SIOCGIFINDEX)?.ifr_mtu as libc::c_int
                },
                sll_hatype: 1,
                sll_pkttype: 0,
                sll_halen: 6,
                sll_addr: [0; 8],
            };

            unsafe {
                let res = libc::bind(
                    self.fd,
                    &sockaddr as *const libc::sockaddr_ll as *const libc::sockaddr,
                    std::mem::size_of::<libc::sockaddr_ll>() as libc::socklen_t,
                );
                if res == -1 {
                    return Err(std::io::Error::last_os_error());
                }
            }

            Ok(())
        }
        #[cfg(not(target_os = "linux"))]
        {
            Err(std::io::Error::new(
                std::io::ErrorKind::Other,
                "Not supported",
            ))
        }
    }

    pub fn interface_mtu(&mut self) -> std::io::Result<usize> {
        #[cfg(target_os = "linux")]
        {
            self.ifreq
                .ioctl(self.fd, SIOCGIFMTU)
                .map(|mtu| unsafe { mtu.ifr_mtu as usize })
        }
        #[cfg(not(target_os = "linux"))]
        {
            Err(std::io::Error::new(
                std::io::ErrorKind::Other,
                "Not supported",
            ))
        }
    }

    pub fn recv(&mut self, buffer: &mut [u8]) -> std::io::Result<usize> {
        let len = unsafe {
            libc::recv(
                self.fd,
                buffer.as_mut_ptr() as *mut libc::c_void,
                buffer.len(),
                0,
            )
        };

        if len == -1 {
            let err = std::io::Error::last_os_error();
            if err.kind() == std::io::ErrorKind::WouldBlock {
                return Err(err);
            } else {
                panic!("err: {:?}", err);
            }
        }
        Ok(len as usize)
    }

    pub fn send(&mut self, buffer: &[u8]) -> std::io::Result<usize> {
        let len = unsafe {
            libc::send(
                self.fd,
                buffer.as_ptr() as *const libc::c_void,
                buffer.len(),
                0,
            )
        };

        if len == -1 {
            let err = std::io::Error::last_os_error();
            if err.kind() == std::io::ErrorKind::WouldBlock {
                return Err(err);
            } else {
                panic!("err: {:?}", err);
            }
        }
        Ok(len as usize)
    }
}

impl Drop for NetDevice {
    fn drop(&mut self) {
        unsafe {
            libc::close(self.fd);
        }
    }
}
